"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var PAREN_PAIRS = '❰❮❬❨❪﹙₍₎﹚❫❩❭❯❱';
var stringifyWithFnName_1 = require("../utils/stringifyWithFnName");
var lazyDefine_1 = require("../utils/lazyDefine");
var Problem = (function () {
    function Problem(val, failsPredicate, subproblems, msg) {
        var _this = this;
        this.isProblem = true;
        if (arguments.length !== 4) {
            throw 'Problem arg len err';
        }
        this.val = val;
        this.name = 'Problem';
        this.failsPredicate = failsPredicate;
        // this.stack = (new Error()).stack;
        this.shortMessage = msg;
        this.subproblems = subproblems;
        lazyDefine_1.default(this, 'message', function () { return _constructMessage(_this, 0); });
        this.toString = function () { return _this.message; };
    }
    ;
    return Problem;
}());
exports.default = Problem;
function _constructMessage(_a, lvl) {
    var subproblems = _a.subproblems, val = _a.val, shortMessage = _a.shortMessage;
    if (Array.isArray(subproblems)) {
        var reasons;
        if (subproblems.length === 0) {
            return shortMessage + "; val: " + stringifyWithFnName_1.default(val, null, 2);
        }
        else {
            reasons = subproblems.map(function (r) {
                return "" + _open(lvl) + _constructMessage(r, lvl + 1) + _close(lvl);
            });
            return shortMessage + ", because\n" + _repeatStr(' ', lvl * 2) + " " + reasons.join(', ');
        }
    }
    else if (typeof subproblems === 'object') {
        reasons = [];
        for (var name in subproblems) {
            if (subproblems.hasOwnProperty(name)) {
                reasons.push("-> " + name + ": " + _open(lvl) + " " + _constructMessage(subproblems[name], lvl + 1) + _close(lvl));
            }
        }
        return shortMessage + ", because\n" + _repeatStr(' ', lvl * 2) + " " + reasons.join(', ');
    }
}
function _repeatStr(str, n) {
    var r = '';
    for (var i = 0; i < n; i += 1) {
        r += str;
    }
    return r;
}
function _open(lvl) {
    return PAREN_PAIRS[lvl];
}
function _close(lvl) {
    return PAREN_PAIRS[PAREN_PAIRS.length - lvl - 1];
}
//# sourceMappingURL=Problem.js.map