"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var resolvedMaps = [];
var namespace_types_1 = require("../clauses/namespace.types");
function resolve(clauseRef, registry) {
    var map = _findFirst(resolvedMaps, function (_a) {
        var registryRef = _a[0], m = _a[1];
        if (registryRef === registry) {
            return m;
        }
    });
    if (!map) {
        map = _createResolveMap(registry);
        resolvedMaps.push([registry, map]);
    }
    return _resolveWithMap(map, clauseRef);
}
exports.default = resolve;
function getDefList(registry) {
    var map = _createResolveMap(registry);
    var groups = {};
    map.forEach(function (curr) {
        var p = curr[0];
        if (!groups[p]) {
            groups[p] = [];
        }
        groups[p].push(curr);
    });
    return groups;
}
exports.getDefList = getDefList;
function _createResolveMap(registry) {
    var r = [];
    var conformedReg = namespace_types_1.NamespaceObjClause.conform(registry);
    _walk(null, null, conformedReg, r);
    return r;
}
function _walk(prefix, currN, creg, r) {
    var currNs = prefix ? prefix + "." + currN : currN;
    var subnamespaces;
    for (var key in creg) {
        if (creg.hasOwnProperty(key)) {
            switch (key) {
                case "subNamespaces":
                    subnamespaces = creg[key];
                    for (var sns in subnamespaces) {
                        if (subnamespaces.hasOwnProperty(sns)) {
                            _walk(currNs, sns, subnamespaces[sns], r);
                        }
                    }
                    break;
                case ".expr":
                    r.push(["" + (prefix ? prefix : ''), "" + currN, creg['.expr']]);
                    break;
                default:
                    break;
            }
        }
    }
}
function _resolveWithMap(map, clauseRef) {
    var path = _findFirst(map, function (_a) {
        var p = _a[0], n = _a[1], r = _a[2];
        if (r === clauseRef ||
            _tryStripPredClause(r) === _tryStripPredClause(clauseRef)) {
            return p + "/" + n;
        }
    });
    return path;
}
function _tryStripPredClause(expr) {
    if (expr.type === 'PRED') {
        return expr.opts.predicate;
    }
    else {
        return expr;
    }
}
function _findFirst(array, fn) {
    for (var i = 0; i < array.length; i++) {
        var r = fn(array[i]);
        if (r) {
            return r;
        }
    }
    return null;
}
//# sourceMappingURL=resolve.js.map