"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Clause_1 = require("../models/Clause");
var isClause_1 = require("../utils/isClause");
var isPred_1 = require("../utils/isPred");
var clauseFromAlts_1 = require("../utils/clauseFromAlts");
var isProblem_1 = require("../utils/isProblem");
var isClauseName_1 = require("../utils/isClauseName");
var namedFn_1 = require("../utils/namedFn");
var isClauseRef_1 = require("../utils/isClauseRef");
var isDelayedClause_1 = require("../utils/isDelayedClause");
var constants_1 = require("../core/constants");
var coerceIntoClause_1 = require("../utils/coerceIntoClause");
var fclause_1 = require("./fclause");
exports.fclause = fclause_1.default;
var walk_1 = require("../walk");
var isObj_1 = require("../preds/isObj");
var isStr_1 = require("../preds/isStr");
var oneOf_1 = require("../preds/oneOf");
var isPlainObj_1 = require("../preds/isPlainObj");
var clauseClause = coerceIntoClause_1.default(isClause_1.default);
var nameClause = coerceIntoClause_1.default(isClauseName_1.default);
var catOp = genMultiArgOp(constants_1.default.CAT);
var orOp = genMultiArgOp(constants_1.default.OR);
var zeroOrMoreOp = genSingleArgOp(constants_1.default.Z_OR_M);
var oneOrMoreOp = genSingleArgOp(constants_1.default.O_OR_M);
var zeroOrOneOp = genSingleArgOp(constants_1.default.Z_OR_O);
var collOfOp = genSingleArgOp(constants_1.default.COLL_OF);
exports.ClauseClause = coerceIntoClause_1.default(isClause_1.default);
exports.ClauseRefClause = coerceIntoClause_1.default(isClauseRef_1.default);
exports.DelayedClauseClause = coerceIntoClause_1.default(isDelayedClause_1.default);
exports.PredClause = coerceIntoClause_1.default(isPred_1.default);
// helper method for constructing labelled structure
function _labelled() {
    var arr = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        arr[_i] = arguments[_i];
    }
    return {
        expressions: {
            withLabels: arr.map(function (_a) {
                var name = _a[0], type = _a[1], v = _a[2];
                return ({ name: name, expr: (_b = {}, _b[type] = v, _b) });
                var _b;
            })
        }
    };
}
function _unlabelled() {
    var arr = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        arr[_i] = arguments[_i];
    }
    return {
        expressions: {
            withoutLabels: arr.map(function (_a) {
                var type = _a[0], v = _a[1];
                return (_b = {}, _b[type] = v, _b);
                var _b;
            })
        }
    };
}
exports.ExprClause = orOp(_labelled(['clause', 'clause', exports.ClauseClause], ['pred', 'clause', exports.PredClause]));
exports.NameExprOptionalComment = catOp(_labelled(['name', 'clause', nameClause], ['comment', 'clause', zeroOrOneOp({ expr: { pred: isStr_1.default } })], ['expr', 'clause', exports.ExprClause]));
exports.MultipleArgClause = catOp(_labelled([
    'expressions', 'clause',
    orOp(_labelled([
        'withLabels', 'clause',
        orOp(_unlabelled([
            'clause',
            zeroOrMoreOp({
                expr: { clause: exports.NameExprOptionalComment },
            })
        ], [
            'clause',
            collOfOp({
                expr: { clause: exports.NameExprOptionalComment },
            })
        ]))
    ], [
        'withoutLabels', 'clause',
        zeroOrMoreOp({
            expr: { clause: exports.ExprClause },
        })
    ]))
], [
    'options', 'clause',
    zeroOrOneOp({
        expr: { pred: isPlainObj_1.default }
    })
]));
function andOp(exprs) {
    var andS = new Clause_1.default({
        type: 'AND',
        exprs: [],
        opts: { conformedExprs: exprs },
        conformFn: function andConform(x) {
            return walk_1.default(andS, x, { conform: true });
        },
        generateFn: null,
    });
    return andS;
}
var multipleArgNoDupeClause = andOp([{ clause: exports.MultipleArgClause },
    { pred: noDupelicateLabels }]);
function noDupelicateLabels(_a) {
    var withLabels = _a.expressions.withLabels;
    if (withLabels) {
        var byFar = [];
        for (var i = 0; i < withLabels.length; i += 1) {
            var lbl = withLabels[i].name;
            if (byFar.indexOf(lbl) >= 0) {
                throw new Error("Duplicate label detected: " + lbl);
            }
            byFar.push(lbl);
        }
    }
    return true;
}
exports.AndFnClause = fclause_1.default({
    args: oneOrMoreOp({ expr: { clause: exports.ExprClause }
    }),
    ret: isClause_1.default,
});
var multipleArgOpClause = {
    args: multipleArgNoDupeClause,
    ret: clauseClause,
};
var singleArgOpClauseFn = function (optClauseAlts) { return ({
    args: catOp(_labelled(['expr', 'clause', exports.ExprClause], [
        'opts', 'clause',
        zeroOrOneOp({ expr: optClauseAlts })
    ])),
    ret: clauseClause,
}); };
function genMultiArgOp(type) {
    return namedFn_1.default(type, function _(_a) {
        var _b = _a.expressions, withLabels = _b.withLabels, withoutLabels = _b.withoutLabels, options = _a.options;
        var exprs;
        if (withLabels) {
            exprs = withLabels;
            var coercedExprs = exprs.map(function (p) {
                var alts = p.expr;
                var s = clauseFromAlts_1.default(alts);
                return Object.assign({}, p, {
                    expr: s,
                    clause: undefined, pred: undefined,
                    clauseRef: undefined, delayedClause: undefined
                });
            });
            var opts = Object.assign({}, options, {
                named: true,
            });
            var s = new Clause_1.default({
                type: type,
                exprs: coercedExprs,
                opts: opts,
                conformFn: function conform(x) {
                    return walk_1.default(s, x, { conform: true });
                },
                generateFn: null,
            });
            return s;
        }
        else if (withoutLabels) {
            exprs = withoutLabels;
            coercedExprs = exprs.map(function (p) {
                var s;
                if (p.clause) {
                    s = p.clause;
                    return Object.assign({}, p, { expr: s, clause: undefined });
                }
                else if (p.pred) {
                    s = coerceIntoClause_1.default(p.pred);
                    return Object.assign({}, p, { expr: s, pred: undefined });
                }
                else if (p.clauseRef) {
                    s = p.clauseRef;
                    return Object.assign({}, p, { expr: s, clauseRef: undefined });
                }
                else if (p.delayedClause) {
                    s = p.delayedClause;
                    return Object.assign({}, p, { expr: s, delayedClause: undefined });
                }
                else {
                    console.error(p);
                    throw '!';
                }
            });
            var opts = Object.assign({}, options, {
                named: false,
            });
            s = new Clause_1.default({
                type: type,
                exprs: coercedExprs,
                opts: opts,
                conformFn: function conform(x) {
                    return walk_1.default(s, x, { conform: true });
                },
                generateFn: null,
            });
            return s;
        }
        else {
            // empty case
            s = new Clause_1.default({
                type: type,
                exprs: [],
                opts: {},
                conformFn: function conform(x) {
                    return walk_1.default(s, x, { conform: true });
                },
                generateFn: null,
            });
            return s;
        }
    });
}
function genSingleArgOp(type) {
    return namedFn_1.default(type, function _(conformedArgs) {
        var p = conformedArgs.expr;
        var opts = conformedArgs.opts;
        var expr;
        if (p.clause) {
            expr = p.clause;
        }
        else if (p.pred) {
            expr = coerceIntoClause_1.default(p.pred);
        }
        else if (p.clauseRef) {
            expr = p.clauseRef;
        }
        else if (p.delayedClause) {
            expr = p.delayedClause;
        }
        else {
            console.error(p);
            throw 'internal err';
        }
        var sureClause = coerceIntoClause_1.default(expr);
        var s = new Clause_1.default({
            type: type,
            exprs: [sureClause],
            opts: Object.assign({}, opts, { enclosedClause: sureClause }),
            conformFn: null,
            generateFn: null,
        });
        s.conform = function conform(x) {
            return walk_1.default(s, x, { conform: true });
        };
        return s;
    });
}
function isPropName(x) {
    return isStr_1.default(x);
}
var TYPE_SHAPE = 'SHAPE';
var TYPE_MAP_OF = 'MAP_OF';
var FieldDefs = mapOfOp({
    keyExpression: {
        clause: coerceIntoClause_1.default(isStr_1.default),
    },
    valExpression: {
        clause: orOp(_labelled(['valExpressionOnly', 'clause', exports.ExprClause], [
            'keyValExprPair', 'clause',
            catOp(_labelled(['keyExpression', 'clause', exports.ExprClause], ['valExpression', 'clause', exports.ExprClause]))
        ]))
    },
});
var KeyOnlyArray = zeroOrMoreOp({
    expr: { pred: isPropName }
});
var KeyArrayOrFieldDefs = orOp(_labelled(['keyList', 'clause', KeyOnlyArray], ['fieldDefs', 'clause', FieldDefs]));
exports.ShapeArgs = shapeOp({
    shapeArgs: {
        optionalFields: {
            opt: {
                fieldDefs: {
                    'requiredFields': {
                        keyValExprPair: {
                            keyExpression: {
                                pred: oneOf_1.default('req', 'required'),
                            },
                            valExpression: {
                                clause: KeyArrayOrFieldDefs
                            },
                        },
                    },
                    'optionalFields': {
                        keyValExprPair: {
                            keyExpression: {
                                pred: oneOf_1.default('opt', 'optional'),
                            },
                            valExpression: {
                                clause: KeyArrayOrFieldDefs
                            },
                        },
                    },
                }
            }
        },
    }
});
exports.MapOfFnClause = fclause_1.default({
    args: catOp(_labelled(['keyExpression', 'clause', exports.ExprClause], ['valExpression', 'clause', exports.ExprClause])),
    ret: isClause_1.default,
});
exports.TestClause = shapeOp({
    shapeArgs: {
        optionalFields: {
            opt: {
                fieldDefs: {
                    'requiredFields': {
                        keyValExprPair: {
                            keyExpression: {
                                pred: oneOf_1.default('req', 'required'),
                            },
                            valExpression: {
                                clause: KeyArrayOrFieldDefs
                            },
                        },
                    },
                    'optionalFields': {
                        keyValExprPair: {
                            keyExpression: {
                                pred: oneOf_1.default('opt', 'optional'),
                            },
                            valExpression: {
                                clause: KeyArrayOrFieldDefs
                            },
                        },
                    },
                }
            }
        },
    }
});
exports.ShapeFnClause = fclause_1.default({
    args: catOp(_labelled(['shapeArgs', 'clause', exports.ShapeArgs])),
    ret: isClause_1.default,
});
function mapOfOp(cargs) {
    if (isProblem_1.default(cargs)) {
        throw cargs;
    }
    var keyExpression = cargs.keyExpression, valExpression = cargs.valExpression;
    var s = new Clause_1.default({
        type: TYPE_MAP_OF,
        exprs: [],
        opts: { keyExpression: keyExpression, valExpression: valExpression },
        conformFn: function mapOfConform(x) {
            return walk_1.default(s, x, { conform: true });
        },
        generateFn: null,
    });
    return s;
}
function shapeOp(cargs) {
    if (isProblem_1.default(cargs)) {
        throw cargs;
    }
    // const { shapeArgs: { requiredFields, optionalFields } } = cargs;
    var s = new Clause_1.default({
        type: TYPE_SHAPE,
        exprs: [],
        opts: { conformedArgs: cargs },
        conformFn: function shapeConform(x) {
            return walk_1.default(s, x, { conform: true });
        },
        generateFn: null,
    });
    return s;
}
exports.shape = exports.ShapeFnClause.instrumentConformed(shapeOp);
exports.keys = exports.shape;
exports.mapOf = exports.MapOfFnClause.instrumentConformed(mapOfOp);
exports.CollOfClause = fclause_1.default(singleArgOpClauseFn({ pred: isObj_1.default }));
exports.collOf = exports.CollOfClause.instrumentConformed(collOfOp);
exports.CatFnClause = fclause_1.default(multipleArgOpClause);
exports.OrFnClause = fclause_1.default(multipleArgOpClause);
exports.ZeroOrMoreFnClause = fclause_1.default(singleArgOpClauseFn({ pred: isObj_1.default }));
exports.OneOrMoreFnClause = fclause_1.default(singleArgOpClauseFn({ pred: isObj_1.default }));
exports.ZeroOrOneFnClause = fclause_1.default(singleArgOpClauseFn({ pred: isObj_1.default }));
exports.and = exports.AndFnClause.instrumentConformed(andOp);
exports.cat = exports.CatFnClause.instrumentConformed(catOp);
exports.or = exports.OrFnClause.instrumentConformed(orOp);
exports.zeroOrMore = exports.ZeroOrMoreFnClause.instrumentConformed(zeroOrMoreOp);
exports.oneOrMore = exports.OneOrMoreFnClause.instrumentConformed(oneOrMoreOp);
exports.zeroOrOne = exports.ZeroOrOneFnClause.instrumentConformed(zeroOrOneOp);
var core = {
    cat: exports.cat, or: exports.or, zeroOrMore: exports.zeroOrMore, oneOrMore: exports.oneOrMore, zeroOrOne: exports.zeroOrOne,
    ExprClause: exports.ExprClause, ClauseClause: exports.ClauseClause, PredClause: exports.PredClause, DelayedClauseClause: exports.DelayedClauseClause, ClauseRefClause: exports.ClauseRefClause,
    CatFnClause: exports.CatFnClause,
    AndFnClause: exports.AndFnClause,
    OrFnClause: exports.OrFnClause,
    ZeroOrMoreFnClause: exports.ZeroOrMoreFnClause, OneOrMoreFnClause: exports.OneOrMoreFnClause, ZeroOrOneFnClause: exports.ZeroOrOneFnClause,
    CollOfClause: exports.CollOfClause,
    collOf: exports.collOf,
    and: exports.and,
    shape: exports.shape,
    keys: exports.shape,
    mapOf: exports.mapOf,
    ShapeFnClause: exports.ShapeFnClause,
    MapOfFnClause: exports.MapOfFnClause,
};
core['alt'] = core.or;
core['*'] = core.zeroOrMore;
core['+'] = core.oneOrMore;
core['?'] = core.zeroOrOne;
exports.default = core;
// // // // //
// var TestClause = shapeOp({
//   shapeArgs: {
//     req: {
//       fieldDefs: {
//         'a': { valExpressionOnly: { pred: isStr } }
//       }
//     }
//   }
// });
// //
// var r = TestClause.conform({a: 's'});
// console.log(r);
// // //
//# sourceMappingURL=regex.js.map