"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var core_1 = require("../core");
var preds_1 = require("../preds");
var utils_1 = require("../utils");
var _1 = require("../");
var clauseFromAlts_1 = require("../utils/clauseFromAlts");
var fnName_1 = require("../utils/fnName");
var stringifyWithFnName_1 = require("../utils/stringifyWithFnName");
var sExpression_1 = require("../utils/sExpression");
var describe_1 = require("../utils/describe");
var match_1 = require("../utils/match");
//     ----'first'-----  --------'second'---------
// fn( (isStr, isBool)+, (isObj | (isNum, isBool)) )
//     -----'first'----  ---second--
//                        -(objOpt)-
// fn( (isStr, isBool)+,    isObj     )
//     -----'first'----  ----second----
//                       ---(showNum)--
// fn( (isStr, isBool)+, isNum, isBool )
// fn( { <required>: [ 'propA': isNum, <propB: beginsWithS>: any ], <optional>: [  ] } )
// cat('first', oneOrMore(cat (isStr, isBool), 'second': {...}  ))
// console.log( conform( sample, [ 'hello', true, 'abc', false, 32, false ] ) );
function AltHeadNode(label, clause, enclosed) {
    this.label = label;
    this.clause = clause;
    this.enclosed = enclosed;
}
var _a = sExpression_1.genClauses(core_1.or('expression', core_1.ExprClause, 'altNode', core_1.and(preds_1.instanceOf(AltHeadNode), _1.default.shape({
    required: {
        clause: _isPivot,
        enclosed: utils_1.delayed(function () { return PartialableParamItemClause; })
    }
})))), PartialableSExprClause = _a[0], PartialableParamItemClause = _a[1];
var synopsis = core_1.fclause({
    args: core_1.cat(core_1.ExprClause, core_1.zeroOrOne(preds_1.isInt), core_1.zeroOrOne(core_1.maybe(preds_1.isFn)))
}).instrument(function synopsis(clause, limit, replacer) {
    if (limit === void 0) { limit = 20; }
    var sExpr = sExpression_1.default(clause);
    var cSExpr = utils_1.conform(sExpression_1.ParamItemClause, sExpr);
    var pivots = _findPivots(cSExpr, replacer);
    var expanded = pivots.reduce(function (cases, pivot) {
        var r = cases.reduce(function (acc, currCase) {
            if (acc.length > limit) {
                return acc;
            }
            else {
                var cases_1 = _expand(currCase, pivot).cases;
                return acc.concat(cases_1);
            }
        }, []);
        return r;
    }, [sExpr]);
    var results = expanded.map(function (cc) { return _describeCase(cc, replacer); });
    return results;
});
function _strFragments(label, cNode, replacer) {
    var result = [];
    if (label) {
        result = result.concat([label, ': ']);
    }
    var _a = _handler(cNode), head = _a.head, params = _a.params;
    if (!head) {
        return [];
    }
    if (replacer) {
        var interceptR = replacer(head);
        if (interceptR) {
            return result.concat(interceptR);
        }
    }
    if (head.type === 'PRED') {
        return ["" + fnName_1.default(head.opts.predicate)];
    }
    var nodeLabel = describe_1.humanReadable(head);
    var commaedParamFrags;
    if (head.type === 'FCLAUSE') {
        var _b = params.unlabelled[0].item.unquotedParamsMap, _c = _b.args, _d = (_c === void 0 ? {} : _c).singleParam, args = _d === void 0 ? null : _d, _e = _b.ret, _f = (_e === void 0 ? {} : _e).singleParam, ret = _f === void 0 ? null : _f;
        return []
            .concat(args ? _fragmentParamAlts(null, args, replacer) : [])
            .concat([' → '])
            .concat(ret ? _fragmentParamAlts(null, ret, replacer) : ['any']);
    }
    else if (head.type === 'CAT') {
        var labelled = params.labelled, unlabelled = params.unlabelled;
        var commaedParamFrags_1 = [];
        if (labelled) {
            var paramFrags = labelled.reduce(function (acc, _a) {
                var label = _a.label, item = _a.item;
                var lblStr = _processLabel(label);
                return acc.concat([
                    [_fragmentParamAlts(lblStr, item, replacer)],
                ]);
            }, []);
            commaedParamFrags_1 = describe_1.interpose(paramFrags, [', ', describe_1.NEW_LINE]);
        }
        else if (unlabelled) {
            var paramFrags = unlabelled.map(function (_a) {
                var item = _a.item;
                return _fragmentParamAlts(null, item, replacer);
            });
            commaedParamFrags_1 = describe_1.interpose(paramFrags, [', ', describe_1.NEW_LINE]);
        }
        return result.concat(['( '])
            .concat(commaedParamFrags_1)
            .concat([' )']);
    }
    else if (head.type === 'OR') {
        var labelled = params.labelled, unlabelled = params.unlabelled;
        var commaedParamFrags_2 = [];
        if (labelled) {
            var paramFrags = labelled.reduce(function (acc, _a) {
                var item = _a.item;
                return acc.concat([
                    [_fragmentParamAlts(null, item, replacer)],
                ]);
            }, []);
            commaedParamFrags_2 = describe_1.interpose(paramFrags, [' | ', describe_1.NEW_LINE]);
        }
        else if (unlabelled) {
            var paramFrags = unlabelled.map(function (_a) {
                var item = _a.item;
                return _fragmentParamAlts(null, item, replacer);
            });
            commaedParamFrags_2 = describe_1.interpose(paramFrags, [' | ', describe_1.NEW_LINE]);
        }
        return result.concat(['{ '])
            .concat(commaedParamFrags_2)
            .concat([' }']);
    }
    else if (head.type === 'Z_OR_M') {
        var item = params.unlabelled[0].item;
        var processed = _fragmentParamAlts(null, item, replacer);
        return ['( '].concat(result)
            .concat([processed, ' )*']);
    }
    else if (head.type === 'O_OR_M') {
        var item = params.unlabelled[0].item;
        var processed = _fragmentParamAlts(null, item, replacer);
        return ['( ']
            .concat(result)
            .concat([processed, ' )+']);
    }
    else if (head.type === 'Z_OR_O') {
        var item = params.unlabelled[0].item;
        var processed = _fragmentParamAlts(null, item, replacer);
        return ['( ']
            .concat(result)
            .concat([processed, ' )?']);
    }
    else if (head.type === 'COLL_OF') {
        var item = params.unlabelled[0].item;
        var processed = _fragmentParamAlts(null, item, replacer);
        return ['[ ']
            .concat(result)
            .concat([processed, ' ]*']);
    }
    else if (head.type === 'ANY') {
        return result.concat(['any']);
    }
    else if (head.type === 'SHAPE') {
        var r = params.unlabelled[0].item.unquotedParamsMap;
        var _g = params.unlabelled[0].item.unquotedParamsMap, _h = _g.required, _j = (_h === void 0 ? { singleParam: {} } : _h).singleParam.quotedParamsMap, required = _j === void 0 ? null : _j, _k = _g.optional, _l = (_k === void 0 ? { singleParam: {} } : _k).singleParam.quotedParamsMap, optional = _l === void 0 ? null : _l;
        var items = [];
        if (required) {
            for (var key in required) {
                var r1 = match_1.default(required[key], {
                    'keyList': function (list) {
                        return ['[ '].concat(describe_1.interpose(list.map(function (i) { return "\"" + i + "\""; }), [', ']))
                            .concat(' ]');
                    },
                    'singleParam': function (p) {
                        return _fragmentParamAlts(null, p, replacer);
                    }
                }, function () {
                    throw '!e';
                });
                if (r1) {
                    items = items.concat([[key + "*: ", r1]]);
                }
            }
        }
        if (optional) {
            for (var key in optional) {
                var r1 = match_1.default(optional[key], {
                    'keyList': function (list) {
                        return ['[ '].concat(describe_1.interpose(list.map(function (i) { return "\"" + i + "\""; }), [', ']))
                            .concat(' ]');
                    },
                    'singleParam': function (p) {
                        return _fragmentParamAlts(null, p, replacer);
                    }
                }, function () {
                    throw '!e';
                });
                if (r1) {
                    items = items.concat([[key + "?: ", r1]]);
                }
            }
        }
        var commaSepartedItems = describe_1.interpose(items, [', ']);
        return result
            .concat(['{ '])
            .concat(commaSepartedItems)
            .concat([' }']);
    }
    else if (head.type === 'AND') {
        // TODO: just a temporary hack that takes the first expression
        // need more design decisions
        return _fragmentParamAlts(label, params.unlabelled[0].item, replacer);
    }
    else if (head.type === 'MAP_OF') {
        var _m = params.unlabelled[0].item.unquotedParamsMap, keyExprAlts = _m.key.singleParam, valExprAlts = _m.val.singleParam;
        var items = []
            .concat(_fragmentParamAlts(null, keyExprAlts, replacer))
            .concat([', '])
            .concat(_fragmentParamAlts(null, valExprAlts, replacer));
        return result
            .concat(['< '])
            .concat(items)
            .concat([' >']);
    }
    else {
        console.error(head);
        throw 'not supported';
    }
}
function _processLabel(_a) {
    var str = _a.str, quoted = _a.quoted;
    if (str) {
        return str;
    }
    else if (quoted) {
        return quoted.value;
    }
}
function _fragmentParamAlts(label, pAlts, replacer) {
    var r = match_1.default(pAlts, {
        'label': _processLabel,
        'sExpression': function (expr) { return _strFragments(label, expr, replacer); },
        'quotedParamsMap': function (o) { return _fragmentParamsObj(o, replacer); },
        'unquotedParamsMap': function (o) { return _fragmentParamsObj(o, replacer); },
        'optionsObj': function (o) { return stringifyWithFnName_1.default(o); },
        'recursive': function (_a) {
            var expression = _a.expression;
            return [
                '<recursive>: ',
                describe_1.humanReadable(expression)
            ];
        }
    }, function (e) {
        console.error(e);
        throw '!s';
    });
    return r;
}
function _fragmentParamsObj(pObj, replacer) {
    var r = ['< ', describe_1.INDENT_IN, describe_1.NEW_LINE,];
    var body = [];
    var keyExprAlts = pObj.key, valExprAlts = pObj.val;
    var keyR = match_1.default(keyExprAlts, {
        'keyList': function (list) {
            return ['[ '].concat(describe_1.interpose(list.map(function (i) { return "\"" + i + "\""; }), [', ']))
                .concat(' ]');
        },
        'singleParam': function (p) {
            return _fragmentParamAlts(null, p, replacer);
        }
    }, function () {
        throw '!e';
    });
    var valR = match_1.default(valExprAlts, {
        'keyList': function (list) {
            return ['[ '].concat(describe_1.interpose(list.map(function (i) { return "\"" + i + "\""; }), [', ']))
                .concat(' ]');
        },
        'singleParam': function (p) {
            return _fragmentParamAlts(null, p, replacer);
        }
    }, function () {
        throw '!e';
    });
    body.push([keyR]);
    body.push([valR]);
    body = describe_1.interpose(body, [', ', describe_1.NEW_LINE]);
    r = r.concat(body).concat([describe_1.INDENT_OUT, describe_1.NEW_LINE, ' >']);
    return r;
}
function _describeCase(c, replacer) {
    var cc = utils_1.conform(PartialableSExprClause, c);
    if (_1.default.isProblem(cc)) {
        throw '!!';
    }
    var fragments = _strFragments(null, cc, replacer);
    var r = describe_1.fragsToStr(fragments, 0, 0);
    return r;
}
function _handler(alts) {
    var headAlts = alts.head, params = alts.params;
    return match_1.default(headAlts, {
        'expression': function (e) { return ({ head: clauseFromAlts_1.default(e), params: params }); },
        'altNode': function (_a) {
            var enclosed = _a.enclosed;
            return match_1.default(enclosed, {
                'sExpression': _handler,
            }, function () { });
        }
    }, function () {
        throw '3';
    });
}
function _expand(currCase, pivot) {
    if (_1.default.conform(PartialableSExprClause, currCase).sExpression) {
        var head = currCase[0], params = currCase.slice(1);
        if (head === pivot) {
            var altCases = _makeAlts(head, params);
            return { found: true, cases: altCases };
        }
        else {
            var _loop_1 = function (i) {
                var _a = _expand(params[i], pivot), found = _a.found, cases = _a.cases;
                if (found) {
                    return { value: {
                            found: found,
                            cases: cases.map(function (c) { return _makeAltCase(c, currCase, i); })
                        } };
                }
            };
            for (var i = 0; i < params.length; i += 1) {
                var state_1 = _loop_1(i);
                if (typeof state_1 === "object")
                    return state_1.value;
            }
        }
    }
    if (currCase[0] instanceof AltHeadNode) {
        var _a = currCase[0], enclosed_1 = _a.enclosed, label_1 = _a.label, clause_1 = _a.clause;
        var _loop_2 = function (i) {
            var _a = _expand(enclosed_1[i], pivot), found = _a.found, cases = _a.cases;
            if (found) {
                return { value: {
                        found: found,
                        cases: cases.map(function (c) {
                            return [new AltHeadNode(label_1, clause_1, _makeAltCase(c, enclosed_1, i - 1))];
                        })
                    } };
            }
        };
        for (var i = 0; i < enclosed_1.length; i += 1) {
            var state_2 = _loop_2(i);
            if (typeof state_2 === "object")
                return state_2.value;
        }
    }
    if (currCase instanceof sExpression_1.QuotedParamsMap ||
        currCase instanceof sExpression_1.UnquotedParamsMap) {
        var _loop_3 = function (key) {
            if (currCase.hasOwnProperty(key)) {
                var val = currCase[key];
                var _a = _expand(val, pivot), found = _a.found, cases = _a.cases;
                if (found) {
                    return { value: {
                            found: found,
                            cases: cases.map(function (c) {
                                return _makeAltCaseMap(c, currCase, key);
                            })
                        } };
                }
            }
        };
        for (var key in currCase) {
            var state_3 = _loop_3(key);
            if (typeof state_3 === "object")
                return state_3.value;
        }
    }
    return { found: false, cases: [currCase] };
}
function _makeAlts(pivot, params) {
    if (pivot.opts.named) {
        return pivot.exprs.map(function (_a, idx) {
            var name = _a.name, expr = _a.expr;
            return [new AltHeadNode(name, pivot, params[idx * 2 + 1])];
        });
    }
    else {
        return pivot.exprs.map(function (e, idx) {
            return [new AltHeadNode(null, pivot, params[idx])];
        });
    }
}
function _makeAltCase(item, sExpression, posInParam) {
    return sExpression.slice(0, posInParam + 1)
        .concat([item])
        .concat(sExpression.slice(posInParam + 2));
}
function _makeAltCaseMap(item, map, key) {
    var r;
    if (map instanceof sExpression_1.QuotedParamsMap) {
        r = new sExpression_1.QuotedParamsMap();
    }
    else if (map instanceof sExpression_1.UnquotedParamsMap) {
        r = new sExpression_1.UnquotedParamsMap();
    }
    Object.assign(r, map);
    r[key] = item;
    return r;
}
function _fold(reducer, _a, init, replacer, inFclause) {
    var sExpression = _a.sExpression, quotedParamsMap = _a.quotedParamsMap, unquotedParamsMap = _a.unquotedParamsMap;
    var r = init;
    if (sExpression) {
        var headAlts = sExpression.head, _b = sExpression.params, _c = _b === void 0 ? {} : _b, _d = _c.labelled, labelled = _d === void 0 ? null : _d, _e = _c.unlabelled, unlabelled = _e === void 0 ? null : _e;
        var head_1 = clauseFromAlts_1.default(headAlts);
        var replaced;
        if (replacer) {
            replaced = replacer(head_1);
            if (replaced) {
                return r;
            }
        }
        r = reducer(r, head_1);
        var items = labelled || unlabelled || [];
        r = items.reduce(function (acc, _a) {
            var item = _a.item;
            if (head_1.type === 'FCLAUSE') {
                return _fold(reducer, item, acc, replacer, true);
            }
            else {
                return _fold(reducer, item, acc, replacer);
            }
        }, r);
    }
    else if (quotedParamsMap || unquotedParamsMap) {
        var m = quotedParamsMap || unquotedParamsMap;
        for (var key in m) {
            if (m.hasOwnProperty(key)) {
                if (!(inFclause && unquotedParamsMap && key === 'ret')) {
                    var singleParam = m[key].singleParam;
                    if (singleParam) {
                        r = _fold(reducer, singleParam, r, replacer, inFclause);
                    }
                }
            }
        }
    }
    return r;
}
// A "pivot" is an "or" clause
function _findPivots(cSExpr, replacer) {
    return _fold(function (acc, item) {
        if (_isPivot(item) && acc.indexOf(item) < 0) {
            return acc.concat([item]);
        }
        else {
            return acc;
        }
    }, cSExpr, [], replacer);
}
function _isPivot(expr) {
    return expr.type === 'OR';
}
exports.default = synopsis;
// // //
// var SampleClause = cat(
//   'first', oneOrMore( cat( isStr, isBool ) ),
//   'second', or(
//     'objOpt', isObj,
//     'showNum', cat( isNum, or( isBool, isObj, isFinite ) )
//   ),
//   'third', shape( {
//     required: {
//       'qw': or( isNum, isObj )
//     }
//   } )
//   );
// var SampleFnClause = fclause( {
//   args: SampleClause
// } );
// import { TestClause } from '../core/regex';
// var r = synopsis( TestClause );
// console.log( r );
//# sourceMappingURL=syntax.js.map