"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var sExpression_1 = require("./sExpression");
var humanReadable_1 = require("./humanReadable");
exports.humanReadable = humanReadable_1.default;
var isStr_1 = require("../preds/isStr");
var isProblem_1 = require("./isProblem");
var match_1 = require("./match");
var clauseFromAlts_1 = require("./clauseFromAlts");
var fnName_1 = require("./fnName");
var stringifyWithFnName_1 = require("../utils/stringifyWithFnName");
var repeat_1 = require("./repeat");
exports.NEW_LINE = function () { };
exports.INDENT_IN = function () { };
exports.INDENT_OUT = function () { };
function describe(expr, replacer, space) {
    var sexpr = sExpression_1.default(expr);
    var cSexpr = sExpression_1.SExpressionClause.conform(sexpr);
    if (isProblem_1.default(cSexpr)) {
        console.error(cSexpr);
        throw new Error('The given expression is not a valid expression.');
    }
    var strFragments = _strFragments(cSexpr, replacer);
    var level = 0;
    var r = fragsToStr(strFragments, level, space);
    return r;
}
function _strFragments(cSExpr, replacer) {
    return strFragments(function (_a) {
        var headAlts = _a.head, params = _a.params;
        return { head: clauseFromAlts_1.default(headAlts), params: params };
    }, cSExpr, replacer);
}
function strFragments(headAltsHandler, cNode, replacer) {
    var _a = headAltsHandler(cNode), head = _a.head, params = _a.params;
    if (!head) {
        return [];
    }
    if (replacer) {
        var interceptR = replacer(head);
        if (interceptR) {
            return interceptR;
        }
    }
    if (head.type === 'PRED') {
        return ["" + fnName_1.default(head.opts.predicate)];
    }
    var label = humanReadable_1.default(head);
    var commaedParamFrags;
    if (params) {
        var labelled = params.labelled, unlabelled = params.unlabelled, keyList = params.keyList;
        if (labelled) {
            var paramFrags = labelled.reduce(function (acc, _a) {
                var label = _a.label, item = _a.item;
                return acc.concat([
                    [_processLabel(label),
                        ', ',
                        _fragmentParamAlts(headAltsHandler, item, replacer)
                    ],
                ]);
            }, []);
            commaedParamFrags = interpose(paramFrags, [', ', exports.NEW_LINE]);
        }
        else if (unlabelled) {
            var paramFrags = unlabelled.map(function (_a) {
                var item = _a.item;
                return _fragmentParamAlts(headAltsHandler, item, replacer);
            });
            commaedParamFrags = interpose(paramFrags, [', ', exports.NEW_LINE]);
        }
        else if (keyList) {
            var paramFrags = keyList;
            commaedParamFrags = interpose(paramFrags, [', ']);
        }
        else {
            // console.error( params );
            // throw '!z';
            commaedParamFrags = [];
        }
    }
    else {
        commaedParamFrags = [];
    }
    return [label, '(']
        .concat(commaedParamFrags.length > 1 ?
        [exports.INDENT_IN, exports.NEW_LINE,] : [commaedParamFrags.length === 0 ? '' : ' '])
        .concat(commaedParamFrags)
        .concat(commaedParamFrags.length > 1 ?
        [exports.INDENT_OUT, exports.NEW_LINE,] : [commaedParamFrags.length === 0 ? '' : ' '])
        .concat([')']);
}
function interpose(arr, interArr) {
    if (arr.length === 0) {
        return arr;
    }
    else {
        return arr.reduce(function (acc, curr, idx) {
            if (idx < arr.length - 1 && !isSpecial(curr)) {
                return acc.concat([curr]).concat(interArr);
            }
            else {
                return acc.concat([curr]);
            }
        }, []);
    }
}
exports.interpose = interpose;
function isSpecial(x) {
    return x === exports.NEW_LINE || x === exports.INDENT_IN || x === exports.INDENT_OUT;
}
exports.isSpecial = isSpecial;
function _fragmentParamAlts(headAltsHandler, pAlts, replacer) {
    var r = match_1.default(pAlts, {
        'label': _processLabel,
        'sExpression': function (expr) { return strFragments(headAltsHandler, expr, replacer); },
        'quotedParamsMap': function (o) { return _fragmentParamsObj(headAltsHandler, o, replacer, true); },
        'unquotedParamsMap': function (o) { return _fragmentParamsObj(headAltsHandler, o, replacer, false); },
        'optionsObj': function (o) { return stringifyWithFnName_1.default(o); },
        'recursive': function (_a) {
            var expression = _a.expression;
            return [
                '<recursive>: ',
                humanReadable_1.default(expression)
            ];
        }
    }, function () {
        throw '!s';
    });
    return r;
}
function _processLabel(_a) {
    var str = _a.str, quoted = _a.quoted;
    if (str) {
        return str;
    }
    else if (quoted) {
        return "\"" + quoted.value + "\"";
    }
}
function _fragmentParamsObj(headAltsHandler, pObj, replacer, quote) {
    var r = ['{', exports.INDENT_IN, exports.NEW_LINE,];
    var body = [];
    for (var label in pObj) {
        if (pObj.hasOwnProperty(label)) {
            var item = [];
            item.push(quote ? "\"" + label + "\": " : "<" + label + ">: ");
            var r1 = match_1.default(pObj[label], {
                'keyList': function (list) {
                    return ['[ '].concat(interpose(list.map(function (i) { return "\"" + i + "\""; }), [', ']))
                        .concat(' ]');
                },
                'singleParam': function (p) {
                    return _fragmentParamAlts(headAltsHandler, p, replacer);
                }
            }, function () {
                throw '!e';
            });
            if (r1) {
                item.push(r1);
                body.push(item);
            }
        }
    }
    body = interpose(body, [', ', exports.NEW_LINE]);
    r = r.concat(body).concat([exports.INDENT_OUT, exports.NEW_LINE, '}']);
    return r;
}
function fragsToStr(frags, level, space) {
    var newLevel = level;
    var justNewLine = false;
    return frags.reduce(function (acc, f) {
        if (justNewLine) {
            justNewLine = false;
            acc = acc.concat(repeat_1.default(space * newLevel, ' ').join(''));
        }
        if (isStr_1.default(f)) {
            return acc.concat(f);
        }
        else if (Array.isArray(f)) {
            return acc.concat(fragsToStr(f, newLevel, space));
        }
        else if (f === exports.NEW_LINE) {
            if (space > 0) {
                justNewLine = true;
                return acc.concat('\n');
            }
        }
        else if (f === exports.INDENT_IN) {
            newLevel += 1;
        }
        else if (f === exports.INDENT_OUT) {
            newLevel -= 1;
        }
        else {
            console.error(f);
            throw '!3';
        }
        return acc;
    }, '');
}
exports.fragsToStr = fragsToStr;
exports.default = describe;
//# sourceMappingURL=describe.js.map