"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var isProblem_1 = require("../utils/isProblem");
var Problem_1 = require("../models/Problem");
var fnName_1 = require("../utils/fnName");
var namedFn_1 = require("../utils/namedFn");
var betterThrow_1 = require("../utils/betterThrow");
function fclauseWalker(clause, walkFn) {
    var _a = clause.opts, argsClause = _a.args, retClause = _a.ret, validateFn = _a.fn;
    return {
        trailblaze: fclauseTrailblaze,
        reconstruct: fclauseReconstruct,
    };
    function fclauseTrailblaze(fn) {
        return fn;
    }
    function fclauseReconstruct(fn, walkOpts) {
        if (fn) {
            var conform = walkOpts.conform, instrument = walkOpts.instrument;
            if (conform && instrument) {
                return instrumentConformed(fn, walkOpts);
            }
            else if (instrument) {
                return _instrument(fn, walkOpts);
            }
        }
        else {
            throw new Error('A function must be clauseified for instrumentation.');
        }
    }
    function _instrument(fn, walkOpts) {
        var fnName = fnName_1.default(fn);
        var instrumentedFn = getInstrumentedFn(fnName, fn);
        var namedClauseedFn = namedFn_1.default(fnName, instrumentedFn, '__instrumented');
        return namedClauseedFn;
    }
    function instrumentConformed(fn, walkOpts) {
        var fnName = fnName_1.default(fn);
        var argConformedFn = getArgConformedInstrumentedFn(fnName, fn);
        var namedArgConformedFn = namedFn_1.default(fnName, argConformedFn, '__conformed');
        return namedArgConformedFn;
    }
    function getInstrumentedFn(fnName, fn) {
        return function __instrumented() {
            var args = Array.prototype.slice.call(arguments);
            var instrumentedArgs = checkArgs(fn, fnName, args);
            var retVal = fn.apply(this, instrumentedArgs);
            var instrumentedRetVal = checkRet(fn, fnName, retVal);
            // TODO optimize
            var conformedArgs = argsClause ?
                walkFn(argsClause, args, { conform: true, instrument: true }) :
                args;
            var conformedRetVal;
            if (retClause) {
                conformedRetVal = walkFn(retClause, retVal, { conform: true, instrument: true });
            }
            else {
                conformedRetVal = retVal;
            }
            checkFnRelation(fnName, fn, validateFn, conformedArgs, conformedRetVal);
            return instrumentedRetVal;
        };
    }
    function checkFnRelation(fnName, fn, validateFn, conformedArgs, retVal) {
        if (validateFn) {
            var r = validateFn.call(null, conformedArgs, retVal);
            if (!r) {
                var p = new Problem_1.default({ args: conformedArgs, ret: retVal }, clause, [], "Function " + fnName + " failed valiation on argument-return value relation");
                betterThrow_1.default(p);
            }
        }
    }
    function checkArgs(fn, fnName, args) {
        var displayFnName = fnName || '<anonymous>';
        if (argsClause) {
            var instrumentedArgs = walkFn(argsClause, args, { phase: 'trailblaze' });
            if (isProblem_1.default(instrumentedArgs)) {
                var p = new Problem_1.default(args, clause, [instrumentedArgs], "Arguments for function " + displayFnName + "() is not valid");
                betterThrow_1.default(p);
            }
            else {
                return walkFn(argsClause, instrumentedArgs, { phase: 'reconstruct', conform: false, instrument: true });
            }
        }
        else {
            return args;
        }
    }
    function checkRet(fn, fnName, retVal) {
        var displayFnName = fnName || '<anonymous>';
        if (retClause) {
            var instrumentedRetVal = walkFn(retClause, retVal, { phase: 'trailblaze' });
            if (isProblem_1.default(instrumentedRetVal)) {
                var p = new Problem_1.default(retVal, clause, [instrumentedRetVal], 'Return value for function ' + displayFnName + '() is not valid');
                betterThrow_1.default(p);
            }
            else {
                var r = walkFn(retClause, instrumentedRetVal, { phase: 'reconstruct', instrument: true, conform: false });
                return r;
            }
        }
        else {
            return retVal;
        }
    }
    function getArgConformedInstrumentedFn(fnName, fn) {
        var displayFnName = fnName || '<anonymous>';
        return function __instrumentConformed() {
            var args = Array.prototype.slice.call(arguments);
            var conformedArgs = walkFn(argsClause, args, { conform: true, instrument: true });
            if (isProblem_1.default(conformedArgs)) {
                var p = new Problem_1.default(args, argsClause, [conformedArgs], "Arguments for function " + displayFnName + " is not valid");
                betterThrow_1.default(p);
            }
            var retVal = fn.call(this, conformedArgs);
            checkRet(fn, fnName, retVal);
            if (validateFn) {
                var conformedRetVal;
                if (retClause) {
                    conformedRetVal = walkFn(retClause, retVal, { conform: true, instrument: true });
                }
                else {
                    conformedRetVal = retVal;
                }
                checkFnRelation(fnName, fn, validateFn, conformedArgs, conformedRetVal);
            }
            return retVal;
        };
    }
}
exports.default = fclauseWalker;
//# sourceMappingURL=fclauseWalker.js.map