"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var isProblem_1 = require("../utils/isProblem");
var isUndefined_1 = require("../preds/isUndefined");
var Problem_1 = require("../models/Problem");
var clauseFromAlts_1 = require("../utils/clauseFromAlts");
function shapeWalker(clause, walkFn) {
    var keyConformer;
    var _a = clause.opts.conformedArgs.shapeArgs, requiredFields = _a.requiredFields, optionalFields = _a.optionalFields;
    var reqClauses, optClauses;
    if (requiredFields) {
        reqClauses = requiredFields.req || requiredFields.required;
    }
    if (optionalFields) {
        optClauses = optionalFields.opt || optionalFields.optional;
    }
    return {
        trailblaze: shapeTrailblaze,
        reconstruct: shapeReconstruct,
    };
    function shapeTrailblaze(x, walkOpts) {
        if (['object', 'function'].indexOf(typeof x) < 0) {
            return new Problem_1.default(x, clause, [], 'Value is not an object');
        }
        if (!keyConformer) {
            // lazy
            keyConformer = _genKeyConformer(reqClauses, optClauses, walkFn, walkOpts);
        }
        var keyConformedR = keyConformer(x);
        if (isProblem_1.default(keyConformedR)) {
            return keyConformedR;
        }
        var problems = [];
        var guide = { val: x, groups: [], singles: [] };
        var reqFieldDefs;
        if (reqClauses) {
            reqFieldDefs = reqClauses.fieldDefs;
        }
        if (reqFieldDefs) {
            processFieldDefs_mut(reqFieldDefs, false);
        }
        var optFieldDefs;
        if (optClauses) {
            optFieldDefs = optClauses.fieldDefs;
        }
        if (optFieldDefs) {
            processFieldDefs_mut(optFieldDefs, true);
        }
        if (problems.length > 0) {
            var problemMap = {};
            var failedNames = [];
            for (var i = 0; i < problems.length; i++) {
                var _a = problems[i], n = _a[0], p = _a[1];
                failedNames.push(n);
                problemMap[n] = p;
            }
            var newP = new Problem_1.default(x, clause, problemMap, 'At least one property failed validation: ' + failedNames.join(', '));
            return newP;
        }
        else {
            return guide;
        }
        function processFieldDefs_mut(fieldDefs, optional) {
            fieldLoop: for (var name in fieldDefs) {
                if (fieldDefs.hasOwnProperty(name)) {
                    var keyValAlts = fieldDefs[name];
                    var _a = getFieldGuide(x, name, keyValAlts, walkFn, walkOpts), noop = _a.noop, problem = _a.problem, singleMatch = _a.singleMatch, groupMatch = _a.groupMatch;
                    if (problem) {
                        problems.push([name, problem]);
                        //TODO: improve this;
                        break fieldLoop;
                    }
                    else if (singleMatch) {
                        guide.singles.push(singleMatch);
                    }
                    else if (groupMatch) {
                        if (groupMatch.matchedKeys.length > 0) {
                            guide.groups.push(groupMatch);
                        }
                    }
                    else if (!noop) {
                        throw '!';
                    }
                }
            }
        }
    }
    function shapeReconstruct(_a, walkOpts) {
        var val = _a.val, singles = _a.singles, groups = _a.groups;
        var conform = { walkOpts: walkOpts };
        var mutate = walkOpts.mutate;
        var conformed;
        if (mutate) {
            conformed = val;
        }
        else {
            conformed = Object.assign({}, val);
        }
        singles.forEach(function (fieldGuide) {
            restoreField_mut(conformed, fieldGuide, walkFn, walkOpts);
        });
        groups.forEach(function (_a) {
            var name = _a.name, matchedKeys = _a.matchedKeys;
            if (conform) {
                conformed[name] = {};
                var keysToDel = [];
                matchedKeys.forEach(function (fieldGuide) {
                    restoreField_mut(conformed[name], fieldGuide, walkFn, walkOpts);
                    keysToDel.push(fieldGuide.key);
                });
                _deleteKeys(conformed, keysToDel);
            }
            else {
                matchedKeys.forEach(function (fieldGuide) {
                    restoreField_mut(conformed, fieldGuide, walkFn, walkOpts);
                });
            }
        });
        return conformed;
    }
}
exports.default = shapeWalker;
function restoreField_mut(x, _a, walkFn, walkOpts) {
    var key = _a.key, clause = _a.clause, guide = _a.guide;
    x[key] = walkFn(clause, guide, walkOpts);
}
function _genKeyConformer(reqClauses, optClause, walkFn, walkOpts) {
    return function tryConformKeys(x) {
        if (reqClauses) {
            var missingKeys = [];
            var fieldDefs = reqClauses.fieldDefs, keyList = reqClauses.keyList;
            var reqNames;
            if (fieldDefs) {
                reqNames = [];
                for (var name in fieldDefs) {
                    if (fieldDefs.hasOwnProperty(name)) {
                        reqNames.push(name);
                    }
                }
            }
            else if (keyList) {
                reqNames = [].concat(keyList);
            }
            else {
                throw 'unsupported';
            }
            for (var i = 0; i < reqNames.length; i++) {
                var reqName = reqNames[i];
                //key clause
                if (fieldDefs && fieldDefs[reqName].keyValExprPair) {
                    var found = false;
                    keyTrav: for (var kk in x) {
                        if (x.hasOwnProperty(kk)) {
                            var rr = _conformNamedOrExpr(kk, fieldDefs[reqName].keyValExprPair.keyExpression, walkFn, walkOpts);
                            if (!isProblem_1.default(rr)) {
                                //found a match
                                found = true;
                                break keyTrav;
                            }
                        }
                    }
                    if (!found) {
                        missingKeys.push(reqName);
                    }
                }
                else if (fieldDefs && fieldDefs[reqName].valExpressionOnly) {
                    //key clause
                    if (x.hasOwnProperty(reqName)) {
                        var rrr = _conformNamedOrExpr(x[reqName], fieldDefs[reqName].valExpressionOnly, walkFn, walkOpts);
                        if (isProblem_1.default(rrr)) {
                            //found a match
                            missingKeys.push(reqName);
                        }
                    }
                    else {
                        missingKeys.push(reqName);
                    }
                }
                else if (keyList) {
                    //plain string key
                    if (x[reqName] === undefined) {
                        missingKeys.push(reqName);
                    }
                }
                else {
                    throw '!';
                }
            }
            if (missingKeys.length > 0) {
                return new Problem_1.default(x, reqClauses, [], 'req: keys required: ' + missingKeys.join(', '));
            }
        }
        return x;
    };
}
function _deleteKeys(subject, keys) {
    for (var i = 0; i < keys.length; i++) {
        delete subject[keys[i]];
    }
}
function getFieldGuide(x, name, keyValAlts, walkFn, walkOpts) {
    var valExpressionOnly = keyValAlts.valExpressionOnly, keyValExprPair = keyValAlts.keyValExprPair;
    if (keyValExprPair) {
        var matchedKeys = [];
        var keyExpression = keyValExprPair.keyExpression, valExpression = keyValExprPair.valExpression;
        keysExamine: for (var k in x) {
            if (x.hasOwnProperty(k)) {
                var keyResult = _conformNamedOrExpr(k, keyExpression, walkFn, walkOpts);
                if (!isProblem_1.default(keyResult)) {
                    // single string char case, where name = 0 and x = ''
                    if (x === x[k]) {
                        continue keysExamine;
                    }
                    var valGuide = _conformNamedOrExpr(x[k], valExpression, walkFn, walkOpts);
                    if (isProblem_1.default(valGuide)) {
                        //TODO: improve
                        return { problem: valGuide };
                    }
                    else {
                        matchedKeys.push({ key: k, clause: clauseFromAlts_1.default(valExpression), guide: valGuide });
                    }
                }
            }
        }
        return { groupMatch: { name: name, matchedKeys: matchedKeys } };
    }
    else if (valExpressionOnly) {
        var v = x[name];
        // single string char case, name = 0
        if (!isUndefined_1.default(v) && x[name] !== x) {
            var g = _conformNamedOrExpr(v, valExpressionOnly, walkFn, walkOpts);
            if (isProblem_1.default(g)) {
                return { problem: g };
            }
            else {
                return { singleMatch: { key: name, clause: clauseFromAlts_1.default(valExpressionOnly), guide: g } };
            }
        }
        else {
            return { noop: true };
        }
    }
    else {
        throw '!!';
    }
}
function _conformNamedOrExpr(x, alts, walkFn, walkOpts) {
    var s = clauseFromAlts_1.default(alts);
    var r = walkFn(s, x, walkOpts);
    return r;
}
//# sourceMappingURL=shapeWalker.js.map